package dk.camelot64.kickc.fragment;

import dk.camelot64.cpufamily6502.CpuClobber;
import dk.camelot64.kickc.asm.AsmProgram;
import dk.camelot64.kickc.fragment.signature.AsmFragmentSignature;
import dk.camelot64.kickc.model.Program;
import dk.camelot64.kickc.model.Registers;
import dk.camelot64.kickc.model.TargetCpu;
import dk.camelot64.kickc.model.statements.StatementSource;
import dk.camelot64.kickc.model.symbols.Label;
import dk.camelot64.kickc.model.symbols.ProgramScope;
import dk.camelot64.kickc.model.symbols.Scope;
import dk.camelot64.kickc.model.symbols.Variable;
import dk.camelot64.kickc.model.types.SymbolType;
import dk.camelot64.kickc.model.values.ConstantInteger;
import dk.camelot64.kickc.model.values.ScopeRef;
import dk.camelot64.kickc.model.values.Value;
import dk.camelot64.kickc.parser.AsmParser;
import dk.camelot64.kickc.parser.KickCParser;
import org.antlr.v4.runtime.tree.ParseTree;
import org.antlr.v4.runtime.tree.TerminalNode;

import java.util.LinkedHashMap;

/**
 * Creates ASM fragment templates by parsing ASM in string form.
 */
public class AsmFragmentTemplateParser {

    /**
     * Creates an ASM fragment from a body.
     *
     * @param signature The ASM fragment signature
     * @param body      The ASM body
     * @param targetCpu The target CPU
     * @return The ASM fragment template
     */
    public static AsmFragmentTemplate parse(AsmFragmentSignature signature, String body, TargetCpu targetCpu) {
        final ParsedAsmBody parsedAsmBody = parseAsm(signature, targetCpu, body);
        final AsmFragmentTemplateImpl asmFragmentTemplate = new AsmFragmentTemplateImpl(signature, targetCpu, parsedAsmBody.bodyAsm, parsedAsmBody.clobber, parsedAsmBody.cycles);
        // assertBodyPrint(signature, body, parsedAsmBody.bodyAsm);
        return asmFragmentTemplate;
    }

    /**
     * Assert that the body can be printed and parsed again - adn result in the same parsed value
     * @param signature The signature
     * @param body The original string body
     * @param parsedBody The parsed body
     */
    private static void assertBodyPrint(AsmFragmentSignature signature, String body, KickCParser.AsmLinesContext parsedBody) {
        final String printedBody = AsmTemplateBodyPrinter.print(parsedBody);
        try {
            final KickCParser.AsmLinesContext reparsed = AsmParser.parseAsm(printedBody, new StatementSource(parsedBody));
            if (!isEqual(reparsed, parsedBody))
                throw new RuntimeException("ASM fragment body mismatch for " + signature.getName() + ".\nACTUAL:  '" + normalize(printedBody) + "'\nEXPECTED:'" + normalize(body) + "'\n");
        } catch (Throwable e) {
            throw new RuntimeException("Error parsing printed ASM fragment body for " + signature.getName() + ".\nPRINTED:  '" + normalize(printedBody) + "'\nORIGINAL:'" + normalize(body) + "'\n");
        }
    }

    private static boolean isEqual(ParseTree ctx1, ParseTree ctx2) {
        if (ctx1 == ctx2) return true;
        if (ctx2 == null || ctx1.getClass() != ctx2.getClass()) return false;
        if (ctx1 instanceof TerminalNode)
            return ctx1.getText().equals(ctx2.getText());
        if (ctx1.getChildCount() != ctx2.getChildCount()) return false;
        for (int i = 0; i < ctx1.getChildCount(); i++) {
            if (!isEqual(ctx1.getChild(i), ctx2.getChild(i))) return false;
        }
        return true;
    }

    private static String normalize(String body) {
        body = body.replace("\r", "");
        body = body.replace("\n", " ");
        body = body.replace("  ", " ");
        while (body.endsWith(" ")) {
            body = body.substring(0, body.length() - 1);
        }
        return body;
    }

    /**
     * Create an inline ASM fragment template.
     *
     * @param signature The ASM fragment signature
     * @param asmBody   The parsed ASM body
     * @param targetCpu The target CPU
     * @return The ASM fragment template
     */
    public static AsmFragmentTemplate inline(AsmFragmentSignature signature, KickCParser.AsmLinesContext asmBody, TargetCpu targetCpu) {
        return new AsmFragmentTemplateImpl(signature, targetCpu, asmBody, null, null);
    }

    /**
     * Create a load/store variable
     *
     * @param name        The name
     * @param type        The type
     * @param scope       The scope
     * @param memoryArea  The memory area (zeropage/main memory)
     * @param dataSegment The data segment (in main memory)
     * @return The new PHI-master variable
     */
    private static Variable createLoadStore(String name, SymbolType type, Scope scope, Variable.MemoryArea memoryArea, String dataSegment) {
        return new Variable(name, Variable.Kind.LOAD_STORE, type, scope, memoryArea, dataSegment, null);
    }

    /**
     * Create a PHI master variable
     *
     * @param name        The name
     * @param type        The type
     * @param scope       The scope
     * @param memoryArea  The memory area (zeropage/main memory)
     * @param dataSegment The data segment (in main memory)
     * @return The new PHI-master variable
     */
    private static Variable createPhiMaster(String name, SymbolType type, Scope scope, Variable.MemoryArea memoryArea, String dataSegment) {
        return new Variable(name, Variable.Kind.PHI_MASTER, type, scope, memoryArea, dataSegment, null);
    }

    /**
     * Result from parsing an ASM string body.
     */
    private static class ParsedAsmBody {
        final KickCParser.AsmLinesContext bodyAsm;
        final AsmFragmentClobber clobber;
        final Double cycles;

        public ParsedAsmBody(KickCParser.AsmLinesContext bodyAsm, AsmFragmentClobber clobber, Double cycles) {
            this.bodyAsm = bodyAsm;
            this.clobber = clobber;
            this.cycles = cycles;
        }
    }

    public static int parseCount = 0;

    /**
     * Initialize the fields that require parsing the ASM (bodyAsm, clobber, cycles).
     */
    private static ParsedAsmBody parseAsm(AsmFragmentSignature signature, TargetCpu targetCpu, String body) {
        String signatureName = signature.getName();

        parseCount++;
        if ((parseCount % 1000) == 0) System.out.println("Parsed " + parseCount + " ASM fragment bodies.");

        // Parse the body ASM
        final KickCParser.AsmLinesContext bodyAsm = AsmParser.parseAsm(body, new StatementSource(signature + ".asm", 1, 1, body, -1, -1));
        // Generate a dummy instance to find clobber & cycles
        ProgramScope scope = new ProgramScope();
        LinkedHashMap<String, Value> bindings = new LinkedHashMap<>();
        {
            Variable master = createPhiMaster("z", SymbolType.BYTE, scope, Variable.MemoryArea.ZEROPAGE_MEMORY, null);
            Variable v1 = Variable.createPhiVersion(master, 1);
            v1.setName("z1");
            Variable v2 = Variable.createPhiVersion(master, 2);
            v2.setName("z2");
            Variable v3 = Variable.createPhiVersion(master, 3);
            v3.setName("z3");
            Variable v4 = Variable.createPhiVersion(master, 4);
            v4.setName("z4");
            Variable v5 = Variable.createPhiVersion(master, 5);
            v5.setName("z5");
            Variable v6 = Variable.createPhiVersion(master, 6);
            v6.setName("z6");
            v1.setAllocation(new Registers.RegisterZpMem(2, 1));
            v2.setAllocation(new Registers.RegisterZpMem(4, 1));
            v3.setAllocation(new Registers.RegisterZpMem(6, 1));
            v4.setAllocation(new Registers.RegisterZpMem(8, 1));
            v5.setAllocation(new Registers.RegisterZpMem(9, 1));
            v6.setAllocation(new Registers.RegisterZpMem(10, 1));
            if (signatureName.contains("z1")) bindings.put("z1", v1);
            if (signatureName.contains("z2")) bindings.put("z2", v2);
            if (signatureName.contains("z3")) bindings.put("z3", v3);
            if (signatureName.contains("z4")) bindings.put("z4", v4);
            if (signatureName.contains("z5")) bindings.put("z5", v5);
            if (signatureName.contains("z6")) bindings.put("z6", v6);
        }
        {
            Variable v1 = createLoadStore("m1", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            Variable v2 = createLoadStore("m2", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            Variable v3 = createLoadStore("m3", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            Variable v4 = createLoadStore("m4", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            Variable v5 = createLoadStore("m5", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            Variable v6 = createLoadStore("m6", SymbolType.BYTE, scope, Variable.MemoryArea.MAIN_MEMORY, null);
            v1.setAllocation(new Registers.RegisterMainMem(v1.getVariableRef(), 1, null, false));
            v2.setAllocation(new Registers.RegisterMainMem(v2.getVariableRef(), 1, null, false));
            v3.setAllocation(new Registers.RegisterMainMem(v3.getVariableRef(), 1, null, false));
            v4.setAllocation(new Registers.RegisterMainMem(v4.getVariableRef(), 1, null, false));
            v5.setAllocation(new Registers.RegisterMainMem(v5.getVariableRef(), 1, null, false));
            v6.setAllocation(new Registers.RegisterMainMem(v6.getVariableRef(), 1, null, false));
            if (signatureName.contains("m1")) bindings.put("m1", v1);
            if (signatureName.contains("m2")) bindings.put("m2", v2);
            if (signatureName.contains("m3")) bindings.put("m3", v3);
            if (signatureName.contains("m4")) bindings.put("m4", v4);
            if (signatureName.contains("m5")) bindings.put("m5", v5);
            if (signatureName.contains("m6")) bindings.put("m6", v6);
        }
        if (signatureName.contains("c1")) bindings.put("c1", new ConstantInteger(310L));
        if (signatureName.contains("c2")) bindings.put("c2", new ConstantInteger(320L));
        if (signatureName.contains("c3")) bindings.put("c3", new ConstantInteger(330L));
        if (signatureName.contains("c4")) bindings.put("c4", new ConstantInteger(340L));
        if (signatureName.contains("c5")) bindings.put("c5", new ConstantInteger(350L));
        if (signatureName.contains("c6")) bindings.put("c6", new ConstantInteger(360L));
        if (signatureName.contains("la1")) bindings.put("la1", new Label("@1", scope, true));
        if (signatureName.startsWith("call_")) bindings.put("la1", new Label("@1", scope, true));
        final AsmFragmentTemplateImpl fragmentTemplate = new AsmFragmentTemplateImpl(signature, targetCpu, bodyAsm, null, null);
        AsmFragmentInstance fragmentInstance =
                new AsmFragmentInstance(new Program(), signature, ScopeRef.ROOT, fragmentTemplate, bindings);
        AsmProgram asm = new AsmProgram(targetCpu);
        asm.startChunk(ScopeRef.ROOT, null, signature.getName());
        fragmentInstance.generate(asm);
        CpuClobber cpuClobber = asm.getClobber();
        asm.addStash();
        return new ParsedAsmBody(bodyAsm, new AsmFragmentClobber(cpuClobber), asm.getCycles());
    }

    /**
     * Fix all newlines in the body.
     * - Removes all '\r'
     * - Removes all trailing newlines
     *
     * @param body The body
     * @return The body with fixed newlines
     */
    public static String fixNewlines(String body) {
        body = body.replace("\r", "");
        while (body.length() > 0 && body.charAt(body.length() - 1) == '\n') {
            body = body.substring(0, body.length() - 1);
        }
        return body;
    }
}
