package dk.camelot64.kickc.fragment.synthesis;

import dk.camelot64.kickc.CompileLog;
import dk.camelot64.kickc.fragment.AsmFragmentClobber;
import dk.camelot64.kickc.fragment.AsmFragmentTemplate;
import dk.camelot64.kickc.fragment.AsmTemplateBodyPrinter;
import dk.camelot64.kickc.fragment.signature.AsmFragmentSignature;

import java.util.List;

/**
 * A successfully synthesized or loaded fragment.
 */
public class AsmFragmentSynthesisResult {

    /**
     * The ASM fragment template.
     */
    final AsmFragmentTemplate fragment;

    /**
     * true if the fragment was loaded from disk.
     */
    private final boolean file;
    /**
     * true if the fragment was loaded from the disk cache.
     */
    private final boolean cache;

    /**
     * The synthesis that created the fragment. null if the fragment template was loaded.
     */
    private final AsmFragmentTemplateSynthesisRule synthesis;
    /**
     * The sub fragment template that the synthesis used to create this.
     */
    private final List<AsmFragmentSynthesisResult> subFragments;

    public AsmFragmentSynthesisResult(AsmFragmentTemplate fragment, boolean file, boolean cache, AsmFragmentTemplateSynthesisRule synthesis, List<AsmFragmentSynthesisResult> subFragments) {
        this.fragment = fragment;
        this.file = file;
        this.cache = cache;
        this.synthesis = synthesis;
        this.subFragments = subFragments;
    }

    public boolean isFile() {
        return file;
    }

    public boolean isCache() {
        return cache;
    }

    public AsmFragmentTemplateSynthesisRule getSynthesis() {
        return synthesis;
    }

    public AsmFragmentTemplate getFragment() {
        return fragment;
    }

    public List<AsmFragmentSynthesisResult> getSubFragments() {
        return subFragments;
    }

    public String getSynthesisDescription() {
        StringBuilder name = new StringBuilder();
        name.append(fragment.getSignature());
        if (synthesis != null) {
            name.append(" < ");
            if (subFragments.size() == 1) {
                name.append(subFragments.get(0).getSynthesisDescription());
            } else {
                for (AsmFragmentSynthesisResult subFragment : subFragments) {
                    name.append("(");
                    name.append(subFragment.getSynthesisDescription());
                    name.append(")");
                }
            }
        }
        return name.toString();
    }

    public void logTemplate(CompileLog log, String indent) {
        String prefix;
        if (this.isCache()) {
            prefix = "cached ";
        } else if (this.isFile()) {
            prefix = "loaded ";
        } else {
            prefix = "synthesized ";
        }
        log.append(indent + prefix + this.getSynthesisDescription() + " - clobber:" + fragment.getClobber().toString() + " cycles:" + fragment.getCycles());
        log.append(indent + "  " + getBodyString().replace("\n", "\n" + indent + "  "));
    }

    public AsmFragmentClobber getClobber() {
        return fragment.getClobber();
    }

    public double getCycles() {
        return fragment.getCycles();
    }

    public String getBodyString() {
        return AsmTemplateBodyPrinter.print(fragment.getBodyAsm());
    }

    public AsmFragmentSignature getSignature() {
        return fragment.getSignature();
    }

}
